#include "stdafx.h"

#include "FrameCache.h"

#include "exception.h"

#include <algorithm>
#include <functional>
#include <fstream>

LONG align(LONG n, LONG alignment)
{
	if (alignment == 0) return n;
	return (n + (alignment - 1)) &~ (alignment - 1);
}

// Frame

Frame::Frame(const BITMAPINFOHEADER& header, int width, int height)
{
	DWORD size = align(width * header.biBitCount / 8, 4) * height;
	m_data.resize(
		sizeof(BITMAPINFOHEADER) + 
		header.biClrUsed * sizeof(RGBQUAD) + 
		size);

	LPBITMAPINFO pbi = reinterpret_cast<LPBITMAPINFO>(&m_data[0]);
	pbi->bmiHeader = header;
	pbi->bmiHeader.biWidth = width;
	pbi->bmiHeader.biHeight = height;
	pbi->bmiHeader.biSizeImage = size;

	const RGBQUAD* colors = reinterpret_cast<const BITMAPINFO*>(&header)->bmiColors;
	std::copy(colors, colors + header.biClrUsed, pbi->bmiColors);
}

Frame::Frame(LPVOID pData)
{
	LPBITMAPINFOHEADER pbih = reinterpret_cast<LPBITMAPINFOHEADER>(pData);
	LPBYTE p = reinterpret_cast<LPBYTE>(pData);
	m_data.assign(p, p + pbih->biSize + pbih->biSizeImage);
}

Frame::Frame(const BITMAPINFOHEADER& header, const Bitmap& bitmap)
{
	m_data.resize(
		sizeof(BITMAPINFOHEADER) + 
		256 * sizeof(RGBQUAD));

	LPBITMAPINFO pbi = reinterpret_cast<LPBITMAPINFO>(&m_data[0]);
	pbi->bmiHeader = header;
	pbi->bmiHeader.biWidth = bitmap.width();
	pbi->bmiHeader.biHeight = bitmap.height();
	DWORD size = pbi->bmiHeader.biSizeImage = 
		align(pbi->bmiHeader.biWidth * pbi->bmiHeader.biBitCount / 8, 4) * pbi->bmiHeader.biHeight;

	CDC dc;
	dc.CreateCompatibleDC(0);

	ASSERT_NE(GetDIBits(dc, bitmap.handle(), 0, bitmap.height(), 0,      pbi, DIB_RGB_COLORS), 0, "GetDIBits");
	m_data.resize(
		sizeof(BITMAPINFOHEADER) + 
		pbi->bmiHeader.biClrUsed * sizeof(RGBQUAD) + 
		(std::max)(pbi->bmiHeader.biSizeImage, size)); // better safe than sorry?
	pbi = reinterpret_cast<LPBITMAPINFO>(&m_data[0]);
	ASSERT_NE(GetDIBits(dc, bitmap.handle(), 0, bitmap.height(), data(), pbi, DIB_RGB_COLORS), 0, "GetDIBits");
}


LPBITMAPINFOHEADER Frame::header() const
{
	if (m_data.empty()) return 0;
	return const_cast<LPBITMAPINFOHEADER>(
		reinterpret_cast<const BITMAPINFOHEADER*>(&m_data[0]));
}

LPVOID Frame::data() const
{
	if (m_data.empty()) return 0;
	return const_cast<LPVOID>(
		reinterpret_cast<LPCVOID>(&m_data[header()->biSize + header()->biClrUsed * sizeof(RGBQUAD)]));
}

void Frame::dump() const
{
	std::ofstream ofs("dump", std::ios::out | std::ios::binary);
	ofs.write(reinterpret_cast<const char*>(&m_data[0]), std::streamsize(m_data.size()));
}


// PFrame

void SeparateFields::operator()(BITMAPINFOHEADER& bih) const
{
	bih.biWidth = 2 * align(bih.biWidth, 32 / bih.biBitCount);
	bih.biHeight /= 2;
}


PFrame::PFrame(Frame* frame /*= 0*/)
: Base(frame)
{
	if (*this)
	{
		m_header = *get()->header();
	}
	else
	{
		ZeroMemory(&m_header, sizeof(BITMAPINFOHEADER));
		m_header.biSize = sizeof(BITMAPINFOHEADER);
	}
}

PFrame::PFrame(const PFrame& other)
: Base(other), m_header(other.m_header)
{
}

void PFrame::swap(PFrame& other)
{
	Base::swap(other);
	std::swap(m_header, other.m_header);
}

PFrame& PFrame::operator=(const PFrame& other)
{
	PFrame(other).swap(*this);
	return *this;
}

PFrame PFrame::clone() const
{
	return PFrame(new Frame((*this)->header()));
}

const BITMAPINFOHEADER* PFrame::header() const
{
	return &m_header;
}

LPBITMAPINFOHEADER PFrame::header()
{
	return &m_header;
}

LPVOID PFrame::data() const
{
	return get()->data();
}

LONG PFrame::width() const
{
	return m_header.biWidth;
}

LONG PFrame::height() const
{
	return m_header.biHeight;
}

CSize PFrame::size() const
{
	return CSize(width(), height());
}

CRect PFrame::rect() const
{
	return CRect(CPoint(0, 0), size());
}

CPoint PFrame::fieldOrigin(LONG order) const
{
	return CPoint(order * m_header.biWidth / 2, 0);
}

SIZE_T PFrame::fullLineSize() const
{
	return align(m_header.biWidth * m_header.biBitCount / 8, 4);
}

int PFrame::lineFromY(int y) const
{
	return m_header.biHeight < 0 ? y : (m_header.biHeight - 1 - y);
}

SIZE_T PFrame::byteFromX(int x) const
{
	return x * m_header.biBitCount / 8;
}

LPVOID PFrame::ptr(int x, int y) const
{
	return LPBYTE(data()) + fullLineSize() * lineFromY(y) + byteFromX(x);
}

const RGBTRIPLE& PFrame::at(LONG x, LONG y) const
{
	return *reinterpret_cast<const RGBTRIPLE*>(ptr(x, y));
}
RGBTRIPLE& PFrame::at(LONG x, LONG y)
{
	return *reinterpret_cast<RGBTRIPLE*>(ptr(x, y));
}

void PFrame::copyFrom(const PFrame& source, CRect sourceRect, CRect targetRect)
{
	return copyFrom(source, sourceRect, targetRect, &memcpy);
}

void PFrame::copyFrom(const PBitmap& source)
{
	//CDC dc;
	//dc.CreateCompatibleDC(0);

	//BITMAPINFO bi;
	//bi.bmiHeader = *header();

	//ASSERT_NE(GetDIBits(dc, source->handle(), 0, height(), data(), &bi, DIB_RGB_COLORS), 0, "GetDIBits");

	PFrame sourceFrame(new Frame(*header(), *source));
	copyFrom(sourceFrame, sourceFrame.rect(), sourceFrame.rect());
}



// AviFile

AviFile::AviFile()
	: m_pAviFile(0), m_pAviStream(0), m_pGetFrame(0), 
	m_length(0), m_millifps(0)
{
}

AviFile::~AviFile()
{
	close();
}

void AviFile::open(LPCTSTR lpszFilename)
{
	close();

	ASSERT_SUCCESS(AVIFileOpen(&m_pAviFile, lpszFilename, OF_READ | OF_SHARE_DENY_WRITE, 0), "AVIFileOpen");
	ASSERT_SUCCESS(AVIFileGetStream(m_pAviFile, &m_pAviStream, streamtypeVIDEO, 0), "AVIFileGetStream");

	LONG formatSize = 0;
	ASSERT_SUCCESS(AVIStreamFormatSize(m_pAviStream, 0, &formatSize), "AVIStreamFormatSize");
	if (formatSize < sizeof(BITMAPINFOHEADER)) throw std::runtime_error("AviFile::open: Frame format is too small");

	std::vector<BYTE> format(formatSize, 0);
	ASSERT_SUCCESS(AVIStreamReadFormat(m_pAviStream, 0, &format[0], &formatSize), "AVIStreamReadFormat");

	LPBITMAPINFOHEADER pHeader = reinterpret_cast<LPBITMAPINFOHEADER>(&format[0]);
	pHeader->biBitCount = 24;
	pHeader->biCompression = BI_RGB;
	pHeader->biSizeImage = align(pHeader->biWidth * pHeader->biBitCount / 8, 4) * pHeader->biHeight;

	m_pGetFrame = ASSERT_NE(AVIStreamGetFrameOpen(m_pAviStream, pHeader), 
		PGETFRAME(), "AVIStreamGetFrameOpen");
	m_length = ASSERT_NE(AVIStreamLength(m_pAviStream), -1, "AVIStreamLength");

	AVIFILEINFO fi;
	ASSERT_SUCCESS(AVIFileInfo(m_pAviFile, &fi, sizeof(fi)), "AVIFileInfo");
	m_millifps = fi.dwRate * 1000 / fi.dwScale;
	m_width = fi.dwWidth;
	m_height = fi.dwHeight;
}

void AviFile::close()
{
	m_length   = 0;
	m_millifps = 0;
	m_width    = 0;
	m_height   = 0;
	m_frames.clear();
	if (m_pGetFrame)  { AVIStreamGetFrameClose(m_pGetFrame); m_pGetFrame = 0; }
	if (m_pAviStream) { AVIStreamRelease(m_pAviStream); m_pAviStream = 0; }
	if (m_pAviFile)   { AVIFileRelease(m_pAviFile); m_pAviFile = 0; }
}

PFrame AviFile::operator[](LONG frame) const
{
	Frames::const_iterator it = m_frames.find(frame);
	if (it != m_frames.end()) return it->second;
	if (!m_pGetFrame) return PFrame();
	PFrame result(new Frame(AVIStreamGetFrame(m_pGetFrame, frame)));
	if (result) m_frames.insert(std::make_pair(frame, result));
	return result;
}

PFrame AviFile::getUncached(LONG frame) const
{
	if (!m_pGetFrame) return PFrame();
	return PFrame(new Frame(AVIStreamGetFrame(m_pGetFrame, frame)));
}

void AviFile::erase(LONG frame)
{
	m_frames.erase(frame);
}

LONG AviFile::size() const
{
	return m_length;
}

LONG AviFile::millifps() const
{
	return m_millifps;
}

LONG AviFile::width() const
{
	return m_width;
}

LONG AviFile::height() const
{
	return m_height;
}

CRect AviFile::frameSize() const
{
	return CRect(0, 0, width(), height());
}

Bitmap::Bitmap(LPCSTR lpszFileName)
{
	ASSERT_SUCCESS(m_image.Load(lpszFileName), "CImage::Load");
}

HBITMAP Bitmap::handle() const
{
	//return m_handle;
	return m_image;
}

CSize Bitmap::size() const
{
	//CSize result;
	//ASSERT_NE(GetBitmapDimensionEx(m_handle, &result), 0, "GetBitmapDimensionEx");
	//return result;
	return CSize(width(), height());
}

LONG Bitmap::width() const
{
	//return size().cx;
	return m_image.GetWidth();
}

LONG Bitmap::height() const
{
	//return size().cy;
	return m_image.GetHeight();
}

CRect Bitmap::rect() const
{
	return CRect(CPoint(0, 0), size());
}


PBitmap ImageCache::operator[](const std::string& fileName) const
{
	Images::const_iterator it = m_images.find(fileName);
	if (m_images.end() != it) return it->second;

	PBitmap result(new Bitmap(fileName.c_str()));
	m_images.insert(std::make_pair(fileName, result));

	return result;
}
