#include "stdafx.h"
#include "FrameOperation.h"

#include "FramePatch.h"
#include "FieldPatch.h"
#include "FrameImage.h"
#include "FieldRebuild.h"
#include "MotionInterpolate.h"

#include "vclipDoc.h"

#include <utility>

// FrameOperation

FrameOperation::FrameOperation(CVclipDoc& owner, LONG targetFrame)
: m_pOwner(&owner), m_targetFrame(targetFrame)
{
}

FrameOperation::~FrameOperation()
{
}

void FrameOperation::onModified() const
{
	m_pOwner->SetModifiedFlag();
	m_pOwner->UpdateAllViews(0, m_targetFrame);
}

void FrameOperation::setTargetFrame(LONG targetFrame)
{
	m_targetFrame = targetFrame;
}

LONG FrameOperation::targetFrame() const
{
	return m_targetFrame;
}

// PatchOperation

PatchOperation::PatchOperation(CVclipDoc& owner, LONG targetFrame /*= 0*/, 
	int targetX /*= 0*/, int targetY /*= 0*/,
	int targetWidth /*= 0*/, int targetHeight /*= 0*/)
: FrameOperation(owner, targetFrame), 
	m_targetOrigin(targetX, targetY),
	m_targetSize(targetWidth, targetHeight)
{
}

CPoint PatchOperation::targetOrigin() const
{
	return m_targetOrigin;
}

CRect PatchOperation::targetRect() const
{
	return CRect(m_targetOrigin, m_targetSize);
}

int PatchOperation::targetX() const
{
	return m_targetOrigin.x;
}

int PatchOperation::targetY() const
{
	return m_targetOrigin.y;
}

void PatchOperation::resizeTarget(CSize newSize)
{
	m_targetSize = newSize;
	onModified();
}

CSize PatchOperation::targetSize() const
{
	return m_targetSize;
}

int PatchOperation::targetWidth() const
{
	return m_targetSize.cx;
}

int PatchOperation::targetHeight() const
{
	return m_targetSize.cy;
}

void PatchOperation::moveTargetTo(CPoint newOrigin)
{
	m_targetOrigin = newOrigin;
	onModified();
}

void PatchOperation::setTargetRect(const CRect& rect)
{
	m_targetOrigin = rect.TopLeft();
	m_targetSize = rect.Size();
	onModified();
}

// PatchOperationWithSource

PatchOperationWithSource::PatchOperationWithSource(CVclipDoc& owner,
	LONG sourceFrame /*= 0*/, 
	int sourceX /*= 0*/, int sourceY /*= 0*/,
	int sourceWidth /*= 0*/, int sourceHeight /*= 0*/,
	LONG targetFrame /*= 0*/, 
	int targetX /*= 0*/, int targetY /*= 0*/,
	int targetWidth /*= 0*/, int targetHeight /*= 0*/)
: PatchOperation(owner, targetFrame, targetX, targetY, targetWidth, targetHeight),
	m_sourceFrame(sourceFrame),
	m_sourceOrigin(sourceX, sourceY),
	m_sourceSize(sourceWidth, sourceHeight)
{
}

LONG PatchOperationWithSource::sourceFrame() const
{
	return m_sourceFrame;
}

void PatchOperationWithSource::setSourceFrame(LONG sourceFrame)
{
	m_sourceFrame = sourceFrame;
	onModified();
}

CPoint PatchOperationWithSource::sourceOrigin() const
{
	return m_sourceOrigin;
}

CRect PatchOperationWithSource::sourceRect() const
{
	return CRect(m_sourceOrigin, m_sourceSize);
}

int PatchOperationWithSource::sourceX() const
{
	return m_sourceOrigin.x;
}

int PatchOperationWithSource::sourceY() const
{
	return m_sourceOrigin.y;
}

void PatchOperationWithSource::resizeSource(CSize newSize)
{
	m_sourceSize = newSize;
	onModified();
}

CSize PatchOperationWithSource::sourceSize() const
{
	return m_sourceSize;
}

int PatchOperationWithSource::sourceWidth() const
{
	return m_sourceSize.cx;
}

int PatchOperationWithSource::sourceHeight() const
{
	return m_sourceSize.cy;
}

void PatchOperationWithSource::moveSourceTo(CPoint newOrigin)
{
	m_sourceOrigin = newOrigin;
	onModified();
}

// PFrameOperation

void PFrameOperation::accept(IOperationVisitor& visitor) const
{
	if (get()) get()->accept(visitor);
}

std::ostream& operator<<(std::ostream& stream, const PFrameOperation& op)
{
	the<FrameOperationFactory>().write(stream, op);
	return stream;
}

// FrameOperationFactory

FrameOperationFactory::FrameOperationFactory()
{
}

/*static*/ FrameOperationFactory& FrameOperationFactory::instance()
{
	static FrameOperationFactory theInstance;
	return theInstance;
}

void FrameOperationFactory::insert(const std::string& typeCode, const ICreateFrameOperation& creator)
{
	m_creators.insert(std::make_pair(typeCode, &creator));
}

PFrameOperation FrameOperationFactory::create(const std::string& typeCode, CVclipDoc& owner) const
{
	Creators::const_iterator it = m_creators.find(typeCode);
	if (it == m_creators.end()) return PFrameOperation();
	return (*it->second)(owner);
}

void FrameOperationFactory::read(std::istream& stream, const std::string& typeCode,
	const PFrameOperation& result) const
{
	Creators::const_iterator it = m_creators.find(typeCode);
	(*it->second).read(stream, result);
}

PFrameOperation FrameOperationFactory::read(std::istream& stream, CVclipDoc& owner) const
{
	if (!stream) return PFrameOperation();

	std::string typeCode;
	stream >> typeCode;
	if (!stream) return PFrameOperation();

	PFrameOperation result = create(typeCode, owner);
	if (result) read(stream, typeCode, result);

	return result;
}

void FrameOperationFactory::write(std::ostream& stream, PFrameOperation pfo) const
{
	pfo->writeTo(stream << pfo->typeCode() << " ");
}

PFrameOperation FrameOperationFactory::createFramePatch(CVclipDoc& owner, 
	LONG sourceFrame, int sourceX, int sourceY, int sourceWidth, int sourceHeight,
	LONG targetFrame, int targetX, int targetY, int targetWidth, int targetHeight,
	HOverlay mode) const
{
	return PFrameOperation(new FramePatch(owner,
		sourceFrame, sourceX, sourceY, sourceWidth, sourceHeight,
		targetFrame, targetX, targetY, targetWidth, targetHeight, mode));
}

PFrameOperation FrameOperationFactory::createFieldPatch(CVclipDoc& owner, 
	LONG sourceFrame, LONG sourceOrder, int sourceX, int sourceY, int sourceWidth, int sourceHeight,
	LONG targetFrame, LONG targetOrder, int targetX, int targetY, int targetWidth, int targetHeight,
	HOverlay mode) const
{
	return PFrameOperation(new FieldPatch(owner,
		sourceFrame, sourceOrder, sourceX, sourceY, sourceWidth, sourceHeight,
		targetFrame, targetOrder, targetX, targetY, targetWidth, targetHeight, mode));
}

PFrameOperation FrameOperationFactory::createFrameImage(CVclipDoc& owner,
	LONG targetFrame, const std::string& fileName) const
{
	return PFrameOperation(new FrameImage(owner, targetFrame, fileName));
}

PFrameOperation FrameOperationFactory::createFieldRebuild(CVclipDoc& owner, 
	LONG targetFrame) const
{
	return PFrameOperation(new FieldRebuild(owner, targetFrame));
}

PFrameOperation FrameOperationFactory::createMotionInterpolate(CVclipDoc& owner, 
	LONG targetFrame) const
{
	return PFrameOperation(new MotionInterpolate(owner, targetFrame));
}
