#ifndef SGI_FUNCTIONAL_INCLUDED
#define SGI_FUNCTIONAL_INCLUDED

namespace sgi
{

template <class T> inline T identity_element(std::plus<T>) {
	return T(0);
}
template <class T> inline T identity_element(std::multiplies<T>) {
	return T(1);
}

template <class OuterF, class InnerG>
class unary_compose : public std::unary_function<
	typename InnerG::argument_type,
	typename OuterF::result_type>
{
protected:
	OuterF outer_;
	InnerG inner_;
public:
	unary_compose(const OuterF& outer, const InnerG& inner)
	: outer_(outer), inner_(inner)
	{
	}
	typename OuterF::result_type
	operator()(const typename InnerG::argument_type& arg) const
	{
		return outer_(inner_(arg));
	}
};

template <class OuterF, class InnerG>
inline unary_compose<OuterF, InnerG>
compose1(const OuterF& outer, const InnerG& inner)
{
	return unary_compose<OuterF, InnerG>(outer, inner);
}

template <class OuterF, class InnerG, class InnerH>
class binary_compose : public std::unary_function<
	typename InnerG::argument_type,
	typename OuterF::result_type>
{
protected:
	OuterF f_;
	InnerG g_;
	InnerH h_;
public:
	binary_compose(const OuterF& f, const InnerG& g, const InnerH& h)
	: f_(f), g_(g), h_(h)
	{
	}
	typename OuterF::result_type
	operator()(const typename InnerG::argument_type& arg) const
	{
		return f_(g_(arg), h_(arg));
	}
};

template <class OuterF, class InnerG, class InnerH>
inline binary_compose<OuterF, InnerG, InnerH>
compose2(const OuterF& f, const InnerG& g,
	const InnerH& h)
{
	return binary_compose<OuterF,InnerG,InnerH>(f, g, h);
}

template <class T>
struct identity : public std::unary_function<T, T>
{
	const T& operator()(const T& x) const
	{
		return x;
	}
};

template <class PairT>
struct select1st : public std::unary_function<PairT, typename PairT::first_type>
{
	const typename PairT::first_type&
	operator()(const PairT& p) const
	{
		return p.first;
	}
};

template <class PairT>
struct select2nd : public std::unary_function<PairT, typename PairT::second_type>
{
	const typename PairT::second_type&
	operator()(const PairT& p) const
	{
		return p.second;
	}
};

template <class Arg1T, class Arg2T>
struct project1st : public std::binary_function<Arg1T, Arg2T, Arg1T>
{
	Arg1T operator()(const Arg1T& x, const Arg2T&) const
	{
		return x;
	}
};

template <class Arg1T, class Arg2T>
struct project2nd : public std::binary_function<Arg1T, Arg2T, Arg2T>
{
	Arg2T operator()(const Arg1T&, const Arg2T& y) const
	{
		return y;
	}
};


template <class ResultT>
struct constant_void_fun
{
	typedef ResultT result_type;
	result_type value_;

	constant_void_fun(const result_type& value)
	: value_(value)
	{
	}

	const result_type& operator()() const
	{
		return value_;
	}
};

template <class ResultT, class ArgumentT>
struct constant_unary_fun
{
	typedef ArgumentT argument_type;
	typedef ResultT result_type;
	result_type value_;

	constant_unary_fun(const result_type& value)
	: value_(value)
	{
	}

	const result_type& operator()(const ArgumentT&) const
	{
		return value_;
	}
};

template <class ResultT, class Arg1T, class Arg2T>
struct constant_binary_fun
{
	typedef Arg1T first_argument_type;
	typedef Arg2T second_argument_type;
	typedef ResultT result_type;
	ResultT value_;

	constant_binary_fun(const ResultT& value)
	: value_(value)
	{
	}

	const result_type& operator()(const Arg1T&, const Arg2T&) const
	{
		return value_;
	}
};

template <class ResultT>
inline constant_void_fun<ResultT>
constant0(const ResultT& value)
{
  return constant_void_fun<ResultT>(value);
}

template <class ResultT>
inline constant_unary_fun<ResultT,ResultT>
constant1(const ResultT& value)
{
  return constant_unary_fun<ResultT,ResultT>(value);
}

template <class ResultT>
inline constant_binary_fun<ResultT,ResultT,ResultT>
constant2(const ResultT& value)
{
  return constant_binary_fun<ResultT,ResultT,ResultT>(value);
}

}


#endif
