// vclipView.cpp : implementation of the CVclipView class
//

#include "stdafx.h"
#include "vclip.h"

#include "vclipDoc.h"
#include "vclipView.h"
#include "GotoFrame.h"
#include "MediaTime.h"

#include "FramePatch.h"
#include "FieldPatch.h"
#include "FrameImage.h"
#include "FieldRebuild.h"
#include "MotionInterpolate.h"

#include "FrameRenderer.h"

#include "Dependency.h"

#include <windows.h>

#include <stdexcept>
#include <exception>
#include "exception.h"

#include <algorithm>
#include "algorithms.h"
#include <functional>
#include <iterator>
#include "compose.h"
#include "sgi_functional.h"
#include "adapters.h"
#include <numeric>
#include <utility>

#include <cstdio>
#include <ios>
#include <sstream>
#include <fstream>

#include ".\vclipview.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

CRect normalizeRect(/*by value*/CRect rect)
{
	rect.NormalizeRect();
	return rect;
}

size_t intervalHitTest(LONG min, LONG max, LONG value)
{
	if (value <  min - 2) return 0;
	if (value <= min + 2) return 1;
	if (value <  max - 2) return 2;
	if (value <= max + 2) return 3;
	return 4;
}

LRESULT rectHitTest(CRect rect, CPoint point)
{
	static const LRESULT results[5][5] =
	{
		{ HTNOWHERE, HTNOWHERE,    HTNOWHERE, HTNOWHERE,     HTNOWHERE },
		{ HTNOWHERE, HTTOPLEFT,    HTTOP,     HTTOPRIGHT,    HTNOWHERE },
		{ HTNOWHERE, HTLEFT,       HTCLIENT,  HTRIGHT,       HTNOWHERE },
		{ HTNOWHERE, HTBOTTOMLEFT, HTBOTTOM,  HTBOTTOMRIGHT, HTNOWHERE },
		{ HTNOWHERE, HTNOWHERE,    HTNOWHERE, HTNOWHERE,     HTNOWHERE }
	};

	return results
		[intervalHitTest(rect.top,  rect.bottom, point.y)]
		[intervalHitTest(rect.left, rect.right,  point.x)];
}

//:View modes

void CVclipView::OnUpdateViewFramemode(CCmdUI *pCmdUI)
{
	pCmdUI->SetRadio(m_viewMode == &frameMode);
}
void CVclipView::OnViewFramemode()
{
	m_viewMode = &frameMode;
	Invalidate();
}

void CVclipView::OnUpdateViewFieldmode(CCmdUI *pCmdUI)
{
	pCmdUI->SetRadio(m_viewMode == &fieldMode);
}
void CVclipView::OnViewFieldmode()
{
	m_viewMode = &fieldMode;
	Invalidate();
}

//#View modes

//:Drag states

class NullDrag : public IDragState
{
public:
	void clipCursor(/*by value*/ CRect rect) const {}
	bool mouseMove(const CPoint& point) const { return false; }
	void mouseUp(const CPoint& point) const {}
	bool canCancel() const { return false; }
	void dragCancel() const {}
};

class PatchMove : public IDragState
{
protected:
	const IManipulator* m_manipulator;
	CRect m_offset;
	CRect m_backup;

public:
	PatchMove(const IManipulator& manipulator, const CPoint& point)
		: m_manipulator(&manipulator), m_offset(manipulator.getRect() - point), m_backup(manipulator.getRect())
	{
	}

	void clipCursor(/*by value*/ CRect rect) const
	{
		m_manipulator->deflateRect(rect, m_offset);
		ClipCursor(rect);
	}

	bool mouseMove(const CPoint& point) const
	{
		m_manipulator->setRect(m_offset + point);
		return true;
	}

	void mouseUp(const CPoint& point) const
	{
		m_manipulator->setRect(m_offset + point);
		ClipCursor(0);
	}

	bool canCancel() const
	{
		return true; 
	}
	void dragCancel() const
	{
		m_manipulator->setRect(m_backup);
		ClipCursor(0);
	}
};

class PatchSideResize : public IDragState
{
private:
	const IManipulator* m_manipulator;
	LONG CPoint::* m_pmPoint;
	LONG CRect::* m_pmRect;
	LONG CRect::* m_pmLow;
	LONG CRect::* m_pmHigh;
	LONG m_offset;
	LONG m_backup;

	void setTarget(LONG value) const
	{
		CRect rect = m_manipulator->getRect();
		rect.*m_pmRect = value;
		m_manipulator->setRect(rect);
	}
public:
	PatchSideResize(const IManipulator& manipulator, CPoint point,
		LONG CPoint::* pmPoint, LONG CRect::* pmRect, LONG CRect::* pmLow, LONG CRect::* pmHigh)
	: m_manipulator(&manipulator), 
		m_pmPoint(pmPoint), m_pmRect(pmRect), m_pmLow(pmLow), m_pmHigh(pmHigh),
		m_offset((manipulator.getRect()).*pmRect - point.*pmPoint),
		m_backup((manipulator.getRect()).*pmRect)
	{}

	void clipCursor(/*by value*/ CRect rect) const
	{
		CRect offset(0, 0, 0, 0);

		if (m_offset < 0) offset.*m_pmLow = -m_offset;
		if (m_offset > 0) offset.*m_pmHigh = -m_offset;

		m_manipulator->deflateRect(rect, offset);

		ClipCursor(rect);
	}

	bool mouseMove(const CPoint& point) const
	{
		setTarget(m_offset + point.*m_pmPoint);
		return true;
	}

	void mouseUp(const CPoint& point) const
	{
		setTarget(m_offset + point.*m_pmPoint);
		ClipCursor(0);
	}

	bool canCancel() const
	{
		return true; 
	}
	void dragCancel() const
	{
		setTarget(m_backup);
		ClipCursor(0);
	}
};

class PatchCornerResize : public IDragState
{
private:
	const IManipulator* m_manipulator;
	LONG CRect::* m_pmx;
	LONG CRect::* m_pmy;
	CPoint m_offset;
	CPoint m_backup;

	void setPoint(const CPoint& point) const
	{
		CRect rect = m_manipulator->getRect();
		rect.*m_pmx = point.x;
		rect.*m_pmy = point.y;
		m_manipulator->setRect(rect);
	}

public:
	PatchCornerResize(const IManipulator& manipulator, CPoint point, LONG CRect::* pmx, LONG CRect::* pmy)
	: m_manipulator(&manipulator),
		m_pmx(pmx), m_pmy(pmy), 
		m_offset(CPoint(manipulator.getRect().*pmx, manipulator.getRect().*pmy) - point), 
		m_backup(CPoint(manipulator.getRect().*pmx, manipulator.getRect().*pmy))
	{}

	void clipCursor(/*by value*/ CRect rect) const
	{
		CRect offset(0, 0, 0, 0);
		if (m_offset.x < 0) offset.left = -m_offset.x;
		if (m_offset.y < 0) offset.top = -m_offset.y;
		if (m_offset.x > 0) offset.right = -m_offset.x;
		if (m_offset.y > 0) offset.bottom = -m_offset.y;

		m_manipulator->deflateRect(rect, offset);

		ClipCursor(rect);
	}

	bool mouseMove(const CPoint& point) const
	{
		setPoint(m_offset + point);
		return true;
	}

	void mouseUp(const CPoint& point) const
	{
		setPoint(m_offset + point);
		ClipCursor(0);
	}

	bool canCancel() const
	{
		return true; 
	}
	void dragCancel() const
	{
		setPoint(m_backup);
		ClipCursor(0);
	}
};

void CVclipView::DragCancel()
{
	m_dragState->dragCancel();
	m_dragState = std::auto_ptr<IDragState>(new NullDrag);
}

void CVclipView::EndDrag(CPoint point)
{
	m_dragState->mouseUp(point);
	m_dragState = std::auto_ptr<IDragState>(new NullDrag);
	m_selection.NormalizeRect();
	Invalidate();
}

//#Drag states

//:Postprocessors

template <typename T>
class bitwise_not : public std::unary_function<T, T>
{
public:
	T operator()(T value) const
	{
		return ~value;
	}
};

class Identity : public IPostProcessor
{
public:
	void operator()(PFrame /*frame*/) const
	{
	}
	const IPostProcessor* select(const IPostProcessor* identity, const IPostProcessor* worker) const
	{
		return worker;
	}
} const identity;

class Invert : public IPostProcessor
{
public:
	void operator()(PFrame frame) const
	{
		BYTE* data = reinterpret_cast<BYTE*>(frame.data());
		std::transform(data, data + frame.header()->biSizeImage,
			data,
			bitwise_not<BYTE>());
	}
	const IPostProcessor* select(const IPostProcessor* identity, const IPostProcessor* worker) const
	{
		return identity;
	}
} const invert;

class Deinterlace : public IPostProcessor
{
private:
	RGBTRIPLE blend(const RGBTRIPLE& a, const RGBTRIPLE& b, const RGBTRIPLE& c) const
	{
		RGBTRIPLE result;
		result.rgbtBlue = (a.rgbtBlue + 2 * b.rgbtBlue + c.rgbtBlue) / 4;
		result.rgbtGreen = (a.rgbtGreen + 2 * b.rgbtGreen + c.rgbtGreen) / 4;
		result.rgbtRed = (a.rgbtRed + 2 * b.rgbtRed + c.rgbtRed) / 4;
		return result;
	}
public:
	void operator()(PFrame frame) const
	{
		PFrame copy = frame.clone();
		for (LONG x = 0; x < frame.width(); ++x)
		{
			frame.at(x, 0) = blend(copy.at(x, 1), copy.at(x, 0), copy.at(x, 1));
		}
		for (LONG y = 1; y + 1 < copy.height(); ++y)
		{
			for (LONG x = 0; x < copy.width(); ++x)
			{
				frame.at(x, y) = blend(copy.at(x, y - 1), copy.at(x, y), copy.at(x, y + 1));
			}
		}
		for (LONG x = 0; x < copy.width(); ++x)
		{
			frame.at(x, copy.height() - 1) = blend(copy.at(x, copy.height() - 2), copy.at(x, copy.height() - 1), copy.at(x, copy.height() - 2));
		}
	}
	const IPostProcessor* select(const IPostProcessor* identity, const IPostProcessor* worker) const
	{
		return identity;
	}
} const deinterlace;

void CVclipView::OnUpdateViewInvert(CCmdUI *pCmdUI)
{
	pCmdUI->SetCheck(m_invert == &invert);
}
void CVclipView::OnViewInvert()
{
	m_invert = m_invert->select(&identity, &invert);
	Invalidate();
}

void CVclipView::OnUpdateViewDeinterlace(CCmdUI *pCmdUI)
{
	pCmdUI->SetCheck(m_deinterlace == &deinterlace);
}
void CVclipView::OnViewDeinterlace()
{
	m_deinterlace = m_deinterlace->select(&identity, &deinterlace);
	Invalidate();
}

//#Postprocessors

//:Drawing

PFrame CVclipView::PrepareFrame(CDC* pDC) const
{
	FrameRenderer renderer(pDC, *GetDocument(), m_aviFile, m_imageCache, m_evenCoords);
	PFrame frame = renderer.render(m_pos);

	(*m_invert)(frame);
	(*m_deinterlace)(frame);

	return frame;
}

void CVclipView::OnUpdateViewAccuraterendering(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_aviFile.size());
	pCmdUI->SetCheck(m_accurateRendering);
}
void CVclipView::OnViewAccuraterendering()
{
	m_accurateRendering = !m_accurateRendering;
	Invalidate();
}

void CVclipView::OnDraw(CDC* pDC)
{
	CVclipDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);
	if (!pDoc)
		return;

	CString context = "drawing";
	try
	{
		if (PFrame frame = PrepareFrame(pDC))
		{
			m_viewMode->draw(pDC, m_drawDib, frame, m_order);
		}

		context.Format("drawing active operation");

		m_manipulator->draw(pDC, m_viewMode);
	}
	catch (const std::exception& e)
	{
		CString error;
		error.Format("An error occured while %s.\n%s", context, e.what());
		AfxMessageBox(error, MB_OK | MB_ICONERROR);
	}
	catch (...)
	{
		CString error;
		error.Format("An unknown error occured while %s.", context);
		AfxMessageBox(error, MB_OK | MB_ICONERROR);
	}
}

//#Drawing

//:Manipulators

class NullManipulator : public IManipulator
{
	void draw(CDC* pDC, const ViewMode* mode) const {}
	bool wantClick(const CPoint& point) const { return false; }
	std::auto_ptr<IDragState> mouseDown(const CPoint& point) const
	{
		return std::auto_ptr<IDragState>(new NullDrag); 
	}
	bool canErase() const { return false; }
	void erase(CVclipDoc& document) const {}
	void accept(IOperationVisitor& visitor) const {}
	bool setCursor(const CPoint& point) const { return false; }
	bool isOperation() const { return false; }
	PFrameOperation operation() const { return PFrameOperation(); }
	CRect getRect() const { return CRect(); }
	void setRect(const CRect& rect) const {}
	void deflateRect(CRect& rect, const CRect& offset) const {}
	bool canMove(int dx, int dy) const { return false; }
	void moveBy(int dx, int dy) const {}
	bool canSize(int dx, int dy) const { return false; }
	void sizeBy(int dx, int dy) const {}
};

class SelectionManipulator : public IManipulator
{
private:
	const ViewMode* m_viewMode;
	CRect* m_pSelection;
	CRect m_frameRect;
public:
	explicit SelectionManipulator(const ViewMode* viewMode, CRect& selection, const CRect& frameRect)
		: m_viewMode(viewMode), m_pSelection(&selection), m_frameRect(frameRect)
	{}

	void draw(CDC* pDC, const ViewMode* mode) const
	{
		CRect frameRect = normalizeRect(*m_pSelection);
		if (frameRect.IsRectEmpty()) return;
		pDC->DrawFocusRect(mode->viewFromFrame(frameRect));
	}

	bool wantClick(const CPoint& point) const
	{
		return !!m_pSelection->PtInRect(point);
	}

	std::auto_ptr<IDragState> mouseDown(const CPoint& point) const
	{
		CRect viewRect = m_viewMode->viewFromFrame(*m_pSelection);
		CPoint viewPoint = m_viewMode->viewFromFrame(point);
		switch (rectHitTest(viewRect, viewPoint))
		{
		#define YK_HITTEST(code, handlerClass, params) \
		case code: return std::auto_ptr<IDragState>(new Patch##handlerClass##Resize params);

			YK_HITTEST(HTTOPLEFT    , Corner, (*this, point, &CRect::left , &CRect::top));
			YK_HITTEST(HTTOPRIGHT   , Corner, (*this, point, &CRect::right, &CRect::top));
			YK_HITTEST(HTBOTTOMLEFT , Corner, (*this, point, &CRect::left , &CRect::bottom));
			YK_HITTEST(HTBOTTOMRIGHT, Corner, (*this, point, &CRect::right, &CRect::bottom));
			YK_HITTEST(HTLEFT       , Side  , (*this, point, &CPoint::x, &CRect::left,   &CRect::left, &CRect::right));
			YK_HITTEST(HTRIGHT      , Side  , (*this, point, &CPoint::x, &CRect::right,  &CRect::left, &CRect::right));
			YK_HITTEST(HTTOP        , Side  , (*this, point, &CPoint::y, &CRect::top,    &CRect::top, &CRect::bottom));
			YK_HITTEST(HTBOTTOM     , Side  , (*this, point, &CPoint::y, &CRect::bottom, &CRect::top, &CRect::bottom));

		#undef YK_HITTEST
		case HTCLIENT:
			return std::auto_ptr<IDragState>(new PatchMove(*this, point));
		default:
			*m_pSelection = CRect(point, point);
			return std::auto_ptr<IDragState>(new PatchCornerResize(*this, point,
				&CRect::right, &CRect::bottom));
		}
	}

	bool canErase() const { return false; }
	void erase(CVclipDoc& /*document*/) const {}
	void accept(IOperationVisitor& visitor) const {}

	bool setCursor(const CPoint& point) const
	{
		CRect viewRect = m_viewMode->viewFromFrame(*m_pSelection);
		CPoint viewPoint = m_viewMode->viewFromFrame(point);
		switch (rectHitTest(viewRect, viewPoint))
		{
		case HTTOPLEFT:     case HTBOTTOMRIGHT:  SetCursor(LoadCursor(0, IDC_SIZENWSE)); return true;
		case HTTOP:         case HTBOTTOM:       SetCursor(LoadCursor(0, IDC_SIZENS));   return true;
		case HTTOPRIGHT:    case HTBOTTOMLEFT:   SetCursor(LoadCursor(0, IDC_SIZENESW)); return true;
		case HTLEFT:        case HTRIGHT:        SetCursor(LoadCursor(0, IDC_SIZEWE));   return true;
		case HTCLIENT:                           SetCursor(LoadCursor(0, IDC_SIZEALL));  return true;
		}
		return false;
	}

	bool isOperation() const { return false; }
	PFrameOperation operation() const { return PFrameOperation(); }

	CRect getRect() const
	{
		return *m_pSelection;
	}
	void setRect(const CRect& rect) const
	{
		*m_pSelection = rect;
	}
	void deflateRect(CRect& rect, const CRect& offset) const
	{
		rect.DeflateRect(-offset.left, -offset.top, offset.right, offset.bottom);
	}

	bool canMove(int dx, int dy) const
	{
		return m_frameRect.PtInRect(m_pSelection->TopLeft() + CPoint(dx, dy)) &&
			m_frameRect.PtInRect(m_pSelection->BottomRight() + CPoint(dx, dy));
	}
	void moveBy(int dx, int dy) const
	{
		m_pSelection->OffsetRect(dx, dy);
	}
	bool canSize(int dx, int dy) const
	{
		return !!m_frameRect.PtInRect(m_pSelection->BottomRight() + CPoint(dx, dy));
	}
	void sizeBy(int dx, int dy) const
	{
		m_pSelection->right += dx;
		m_pSelection->bottom += dy;
		m_pSelection->NormalizeRect();
	}
};

class FullFrameManipulator : public IManipulator
{
private:
	PFrameOperation m_operation;
	CRect m_frameRect;
public:
	explicit FullFrameManipulator(PFrameOperation operation, const CRect& frameRect)
		: m_operation(operation), m_frameRect(frameRect)
	{}

	void draw(CDC* pDC, const ViewMode* mode) const
	{
		pDC->DrawFocusRect(mode->viewFromFrame(m_frameRect));
	}
	bool wantClick(const CPoint& point) const
	{
		return true;
	}
	std::auto_ptr<IDragState> mouseDown(const CPoint& point) const
	{
		return std::auto_ptr<IDragState>(new NullDrag);
	}
	bool canErase() const { return true; }
	void erase(CVclipDoc& document) const { document.erase(m_operation); }
	void accept(IOperationVisitor& visitor) const { m_operation->accept(visitor); }
	bool setCursor(const CPoint& point) const
	{
		SetCursor(LoadCursor(0, IDC_ARROW));
		return true;
	}
	bool isOperation() const { return true; }
	PFrameOperation operation() const { return m_operation; }
	CRect getRect() const { return CRect(); }
	void setRect(const CRect& rect) const {}
	void deflateRect(CRect& rect, const CRect& offset) const {}
	bool canMove(int dx, int dy) const { return false; }
	void moveBy(int dx, int dy) const {}
	bool canSize(int dx, int dy) const { return false; }
	void sizeBy(int dx, int dy) const {}
};

class PatchManipulator : public IManipulator
{
private:
	const ViewMode* m_viewMode;
	PFrameOperation m_operation;
	PatchOperation* m_patch;
	CRect m_frameRect;
public:
	explicit PatchManipulator(const ViewMode* viewMode, PatchOperation& patch, const CRect& frameRect)
		: m_viewMode(viewMode), m_operation(&patch), m_patch(&patch), m_frameRect(frameRect)
	{}
	
	~PatchManipulator()
	{
	}

	void draw(CDC* pDC, const ViewMode* mode) const
	{
		pDC->DrawFocusRect(mode->viewFromFrame(normalizeRect(m_patch->targetRect())));
	}
	bool wantClick(const CPoint& point) const
	{
		return !!m_patch->targetRect().PtInRect(point);
	}
	std::auto_ptr<IDragState> mouseDown(const CPoint& point) const
	{
		//CRect viewRect = m_viewMode->viewFromFrame(m_patch->targetRect());
		//CPoint viewPoint = m_viewMode->viewFromFrame(point);
		//switch (rectHitTest(viewRect, viewPoint))
		//{
		//#define YK_HITTEST(code, handlerClass, params) \
		//case code: return std::auto_ptr<IDragState>(new Patch##handlerClass##Resize params);

		//	YK_HITTEST(HTTOPLEFT    , Corner, (*this, point, &CRect::left , &CRect::top));
		//	YK_HITTEST(HTTOPRIGHT   , Corner, (*this, point, &CRect::right, &CRect::top));
		//	YK_HITTEST(HTBOTTOMLEFT , Corner, (*this, point, &CRect::left , &CRect::bottom));
		//	YK_HITTEST(HTBOTTOMRIGHT, Corner, (*this, point, &CRect::right, &CRect::bottom));
		//	YK_HITTEST(HTLEFT       , Side  , (*this, point, &CPoint::x, &CRect::left,   &CRect::left, &CRect::right));
		//	YK_HITTEST(HTRIGHT      , Side  , (*this, point, &CPoint::x, &CRect::right,  &CRect::left, &CRect::right));
		//	YK_HITTEST(HTTOP        , Side  , (*this, point, &CPoint::y, &CRect::top,    &CRect::top, &CRect::bottom));
		//	YK_HITTEST(HTBOTTOM     , Side  , (*this, point, &CPoint::y, &CRect::bottom, &CRect::top, &CRect::bottom));

		//#undef YK_HITTEST
		//case HTCLIENT:
		//default:
		//	return std::auto_ptr<IDragState>(new PatchMove(*this, point));
		//}
		return std::auto_ptr<IDragState>(new PatchMove(*this, point));
	}
	bool canErase() const { return true; }
	void erase(CVclipDoc& document) const { document.erase(PFrameOperation(m_patch)); }
	void accept(IOperationVisitor& visitor) const { m_patch->accept(visitor); }
	bool setCursor(const CPoint& point) const
	{
		CRect viewRect = m_viewMode->viewFromFrame(m_patch->targetRect());
		CPoint viewPoint = m_viewMode->viewFromFrame(point);
		switch (rectHitTest(viewRect, viewPoint))
		{
		//case HTTOPLEFT:     case HTBOTTOMRIGHT:  SetCursor(LoadCursor(0, IDC_SIZENWSE)); return true;
		//case HTTOP:         case HTBOTTOM:       SetCursor(LoadCursor(0, IDC_SIZENS));   return true;
		//case HTTOPRIGHT:    case HTBOTTOMLEFT:   SetCursor(LoadCursor(0, IDC_SIZENESW)); return true;
		//case HTLEFT:        case HTRIGHT:        SetCursor(LoadCursor(0, IDC_SIZEWE));   return true;
		case HTCLIENT:                           SetCursor(LoadCursor(0, IDC_SIZEALL));  return true;
		}
		return false;
	}
	bool isOperation() const { return true; }
	PFrameOperation operation() const { return PFrameOperation(m_patch); }
	
	CRect getRect() const
	{
		return m_patch->targetRect();
	}
	void setRect(const CRect& rect) const
	{
		m_patch->setTargetRect(rect);
	}
	void deflateRect(CRect& rect, const CRect& offset) const {}

	bool canMove(int dx, int dy) const
	{
		// condition: at least one point must be in the frame

		return !(m_frameRect & (m_patch->targetRect() + CPoint(dx, dy))).IsRectEmpty();
	}
	void moveBy(int dx, int dy) const
	{
		m_patch->moveTargetTo(m_patch->targetOrigin() + CPoint(dx, dy));
	}
	bool canSize(int dx, int dy) const { return false; }
	void sizeBy(int dx, int dy) const {}
};

class MakeManipulator : public IOperationVisitor
{
private:
	std::auto_ptr<IManipulator>* m_manipulator;
	const ViewMode* m_viewMode;
	CRect m_frameRect;
public:
	explicit MakeManipulator(std::auto_ptr<IManipulator>& pManipulator, 
		const ViewMode* viewMode, const CRect& frameRect)
	: m_manipulator(&pManipulator), m_viewMode(viewMode), m_frameRect(frameRect)
	{}

	void operator()(FramePatch& patch)
	{
		*m_manipulator = std::auto_ptr<IManipulator>
			(new PatchManipulator(m_viewMode, patch, m_frameRect));
	}
	void operator()(FieldPatch& patch)
	{
		*m_manipulator = std::auto_ptr<IManipulator>
			(new PatchManipulator(m_viewMode, patch, m_frameRect));
	}
	void operator()(FrameImage& image)
	{
		*m_manipulator = std::auto_ptr<IManipulator>
			(new FullFrameManipulator(PFrameOperation(&image), m_frameRect));
	}
	void operator()(FieldRebuild& rebuild)
	{
		*m_manipulator = std::auto_ptr<IManipulator>
			(new FullFrameManipulator(PFrameOperation(&rebuild), m_frameRect));
	}
	void operator()(MotionInterpolate& inter)
	{
		*m_manipulator = std::auto_ptr<IManipulator>
			(new FullFrameManipulator(PFrameOperation(&inter), m_frameRect));
	}
};

class GetOverlayMode : public IOperationVisitor
{
private:
	HOverlay* m_pMode;
public:
	explicit GetOverlayMode(HOverlay& mode)
		: m_pMode(&mode)
	{}

	void operator()(FramePatch& patch)
	{
		*m_pMode = patch.mode();
	}
	void operator()(FieldPatch& patch)
	{
		*m_pMode = patch.mode();
	}
	void operator()(FrameImage& image) {}
	void operator()(FieldRebuild& rebuild) {}
	void operator()(MotionInterpolate& inter) {}
};

void CVclipView::UnselectOperation()
{
	m_manipulator = std::auto_ptr<IManipulator>
		(new SelectionManipulator(m_viewMode, m_selection, m_aviFile.frameSize()));
	m_historyList.SetCurSel(-1);
	UpdateDependencies();
	Invalidate();
}

int IndexOfDataPtr(const CListBox& listbox, const void* ptr)
{
	for (int result = 0, size = listbox.GetCount(); result < size; ++result)
	{
		if (listbox.GetItemDataPtr(result) == ptr) return result;
	}
	return -1;
}

void CVclipView::SelectOperation(PFrameOperation op)
{
	if (!op)
	{
		UnselectOperation();
	}
	else
	{
		op.accept(MakeManipulator(m_manipulator, m_viewMode, m_aviFile.frameSize()));
		op.accept(GetOverlayMode(m_overlayMode));

		m_historyList.SetCurSel(IndexOfDataPtr(m_historyList, op.get()));

		int i;
		for (i = 0; i < m_historyList.GetCount() && m_historyList.GetItemDataPtr(i) != op.get(); ++i);
		m_historyList.SetCurSel(i);

		UpdateDependencies();

		m_selection.SetRectEmpty();

		Invalidate();
	}
}

class HitTest : public std::unary_function<PFrameOperation, void>, public IOperationVisitor
{
private:
	const ViewMode* m_viewMode;
	LONG m_frame;
	LONG m_order;
	CPoint m_framePoint;
	PFrameOperation* m_operation;
public:
	explicit HitTest(const ViewMode* viewMode, LONG frame, LONG order, 
		const CPoint& framePoint, PFrameOperation& operation)
	: m_viewMode(viewMode), m_frame(frame), m_order(order), 
		m_framePoint(framePoint), m_operation(&operation)
	{
	}

	void operator()(FramePatch& patch)
	{
		if (patch.targetRect().PtInRect(m_framePoint))
		{
			*m_operation = PFrameOperation(&patch);
		}
	}

	void operator()(FieldPatch& patch)
	{
		if (patch.targetRect().PtInRect(m_framePoint) && 
			m_viewMode->visible(patch.targetOrder(), m_order))
		{
			*m_operation = PFrameOperation(&patch);
		}
	}

	void operator()(FrameImage& image)
	{
		*m_operation = PFrameOperation(&image);
	}

	void operator()(FieldRebuild& rebuild)
	{
		*m_operation = PFrameOperation(&rebuild);
	}

	void operator()(MotionInterpolate& inter)
	{
		*m_operation = PFrameOperation(&inter);
	}

	void operator()(const PFrameOperation& op)
	{
		if (op->targetFrame() == m_frame)
		{
			op.accept(*this);
		}
	}
};

void CVclipView::OnLButtonDown(UINT nFlags, CPoint viewPoint)
{
	CView::OnLButtonDown(nFlags, viewPoint);

	if (!m_aviFile.size()) return;

	CRect clip;
	GetClientRect(&clip);
	clip.bottom -= 20;
	clip &= m_viewMode->viewFromFrame(m_aviFile.frameSize());
	++clip.bottom; ++clip.right;

	if (!clip.PtInRect(viewPoint)) return;

	CPoint framePoint = m_viewMode->frameFromView(viewPoint);

	CVclipDoc::range range = GetDocument()->operations(m_pos);

	if (MK_CONTROL & nFlags)
	{
		UnselectOperation();
	}
	else if (m_manipulator->wantClick(framePoint))
	{
		// the current manipulator stays
	}
	else
	{
		UnselectOperation();
		PFrameOperation clickedOperation;
		std::for_each(range.first, range.second,
			HitTest(m_viewMode, m_pos, m_order, framePoint, clickedOperation));
		SelectOperation(clickedOperation);
	}

	ClientToScreen(&clip);
	m_dragState = m_manipulator->mouseDown(framePoint);
	m_dragState->clipCursor(clip);

	Invalidate();
}


void CVclipView::OnMouseMove(UINT nFlags, CPoint viewPoint)
{
	CView::OnMouseMove(nFlags, viewPoint);

	if (m_dragState->mouseMove(m_viewMode->frameFromView(viewPoint)))
	{
		Invalidate();
	}
}

void CVclipView::OnLButtonUp(UINT nFlags, CPoint viewPoint)
{
	CView::OnLButtonUp(nFlags, viewPoint);

	EndDrag(m_viewMode->frameFromView(viewPoint));
}

void CVclipView::OnDragcancel()
{
	if (m_dragState->canCancel())
	{
		DragCancel();
	}
	else
	{
		UnselectOperation();
		m_selection.SetRectEmpty();
	}
}

void CVclipView::OnUpdateEditMoveleft(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_manipulator->canMove(-1, 0));
}
void CVclipView::OnEditMoveleft()
{
	m_manipulator->moveBy(-1, 0);
	Invalidate();
}

void CVclipView::OnUpdateEditMoveright(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_manipulator->canMove(1, 0));
}
void CVclipView::OnEditMoveright()
{
	m_manipulator->moveBy(1, 0);
	Invalidate();
}

void CVclipView::OnUpdateEditMoveup(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_manipulator->canMove(0, -1));
}
void CVclipView::OnEditMoveup()
{
	m_manipulator->moveBy(0, -1);
	Invalidate();
}

void CVclipView::OnUpdateEditMovedown(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_manipulator->canMove(0, 1));
}
void CVclipView::OnEditMovedown()
{
	m_manipulator->moveBy(0, 1);
	Invalidate();
}

void CVclipView::OnUpdateEditSizeleft(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_manipulator->canSize(-1, 0));
}
void CVclipView::OnEditSizeleft()
{
	m_manipulator->sizeBy(-1, 0);
	Invalidate();
}

void CVclipView::OnUpdateEditSizeright(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_manipulator->canSize(1, 0));
}
void CVclipView::OnEditSizeright()
{
	m_manipulator->sizeBy(1, 0);
	Invalidate();
}

void CVclipView::OnUpdateEditSizeup(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_manipulator->canSize(0, -1));
}
void CVclipView::OnEditSizeup()
{
	m_manipulator->sizeBy(0, -1);
	Invalidate();
}

void CVclipView::OnUpdateEditSizedown(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_manipulator->canSize(0, 1));
}
void CVclipView::OnEditSizedown()
{
	m_manipulator->sizeBy(0, 1);
	Invalidate();
}

void CVclipView::OnUpdateEditSelectnext(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_historyList.GetCount());
}
void CVclipView::OnEditSelectnext()
{
	int index = m_historyList.GetCurSel();
	index = LB_ERR == index ? 0 : index + 1 < m_historyList.GetCount() ? index + 1 : index;

	SelectOperation(PFrameOperation(static_cast<FrameOperation*>(
		m_historyList.GetItemDataPtr(index))));
}

void CVclipView::OnUpdateEditSelectprevious(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_historyList.GetCount());
}
void CVclipView::OnEditSelectprevious()
{
	int index = m_historyList.GetCurSel();
	index = LB_ERR == index ? 0 : index - 1 >= 0 ? index - 1 : index;

	SelectOperation(PFrameOperation(static_cast<FrameOperation*>(
		m_historyList.GetItemDataPtr(index))));
}

//#Manipulators

//:Edit

void CVclipView::OnUpdateEditCopy(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(!m_selection.IsRectEmpty());
}
void CVclipView::OnEditCopy()
{
	m_clipboard.copy(m_viewMode, m_pos, m_order, m_selection);
}

void CVclipView::OnUpdateEditPaste(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(!m_clipboard.empty());
}
void CVclipView::OnEditPaste()
{
	YK_TRACED("OnEditPaste");
	DragCancel();
	SelectOperation(m_clipboard.paste(*GetDocument(), m_viewMode, m_pos, m_order, m_overlayMode));
}

void CVclipView::OnUpdateEditDelete(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_manipulator->canErase());
}
void CVclipView::OnEditDelete()
{
	m_manipulator->erase(*GetDocument());
	m_manipulator = std::auto_ptr<IManipulator>(new NullManipulator);
}

void CVclipView::OnUpdateEditSelectall(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_aviFile.size());
}
void CVclipView::OnEditSelectall()
{
	UnselectOperation();
	m_selection = m_aviFile.frameSize();
	Invalidate();
}

void CVclipView::OnUpdateEditInsertimage(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_aviFile.size());
}
void CVclipView::OnEditInsertimage()
{
	CFileDialog openDialog(TRUE, ".bmp", 0, OFN_FILEMUSTEXIST | OFN_HIDEREADONLY | OFN_PATHMUSTEXIST,
		/*1*/"Bitmap Files (*.bmp)|*.bmp|"
		/*2*/"GIF Files (*.gif)|*.gif|"
		/*3*/"JPEG Files (*.jpg)|*.jpg|"
		/*4*/"PNG Files (*.png)|*.png|"
		/*5*/"All Image Files (*.bmp;*.gif;*.jpg;*.png)|*.bmp;*.gif;*.jpg;*.png|"
		/*6*/"All Files (*.*)|*.*||", GetParent());
	openDialog.GetOFN().nFilterIndex = 5;

	if (IDOK == openDialog.DoModal())
	{
		SelectOperation(GetDocument()->add(
			the<FrameOperationFactory>().createFrameImage(*GetDocument(),
				m_pos, openDialog.GetFileName().GetString() ) ));
	}
}

void CVclipView::OnUpdateEditRebuildfromadjacentfields(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_pos >= 1 && m_pos + 1 < m_aviFile.size());
}
void CVclipView::OnEditRebuildfromadjacentfields()
{
	SelectOperation(GetDocument()->add(
		the<FrameOperationFactory>().createFieldRebuild(*GetDocument(), m_pos) ));
}

void CVclipView::OnUpdateEditInterpolatemotion(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_pos >= 1 && m_pos + 1 < m_aviFile.size());
}
void CVclipView::OnEditInterpolatemotion()
{
	SelectOperation(GetDocument()->add(
		the<FrameOperationFactory>().createMotionInterpolate(*GetDocument(), m_pos) ));
}

void CVclipView::OnUpdateEditSnapshotframe(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_aviFile.size());
}
void CVclipView::OnEditSnapshotframe()
{
	CString defaultFilename;
	defaultFilename.Format("%d.bmp", m_pos);
	CFileDialog fd(FALSE, ".bmp", defaultFilename, OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,
		"Bitmap Files (*.bmp)|*.bmp|"
		"All Files (*.*)|*.*||", GetParent());
	if (IDOK != fd.DoModal()) return;

	PFrame frame = PrepareFrame(GetDC());

	std::ofstream ofs;
	ofs.exceptions(std::ios::failbit);
	ofs.open(fd.GetFileName(), std::ios_base::out | std::ios_base::binary);

	BITMAPFILEHEADER bfh;
	bfh.bfType = 0x4D42;
	bfh.bfSize = 
		sizeof(BITMAPFILEHEADER) +
		sizeof(BITMAPINFOHEADER) +
		frame.header()->biSizeImage;
	bfh.bfReserved1 = bfh.bfReserved2 = 0;
	bfh.bfOffBits = 
		sizeof(BITMAPFILEHEADER) +
		sizeof(BITMAPINFOHEADER);

	ofs.write(reinterpret_cast<const char*>(&bfh),           sizeof(bfh));
	ofs.write(reinterpret_cast<const char*>(frame.header()), frame.header()->biSize);
	ofs.write(reinterpret_cast<const char*>(frame.data()),   frame.header()->biSizeImage);
}

//#Edit

// CVclipView

IMPLEMENT_DYNCREATE(CVclipView, CView)

BEGIN_MESSAGE_MAP(CVclipView, CView)
	ON_COMMAND(ID_VIEW_OPENVIDEO, OnViewOpenvideo)
	ON_COMMAND(ID_VIEW_NEXTFRAME, OnViewNextframe)
	ON_UPDATE_COMMAND_UI(ID_VIEW_NEXTFRAME, OnUpdateViewNextframe)
	ON_UPDATE_COMMAND_UI(ID_VIEW_PREVIOUSFRAME, OnUpdateViewPreviousframe)
	ON_COMMAND(ID_VIEW_PREVIOUSFRAME, OnViewPreviousframe)
	ON_WM_LBUTTONDOWN()
	ON_WM_LBUTTONUP()
	ON_WM_MOUSEMOVE()
	ON_UPDATE_COMMAND_UI(ID_EDIT_COPY, OnUpdateEditCopy)
	ON_COMMAND(ID_EDIT_COPY, OnEditCopy)
	ON_UPDATE_COMMAND_UI(ID_EDIT_PASTE, OnUpdateEditPaste)
	ON_COMMAND(ID_EDIT_PASTE, OnEditPaste)
	ON_COMMAND(ID_DRAGCANCEL, OnDragcancel)
	ON_UPDATE_COMMAND_UI(ID_EDIT_DELETE, OnUpdateEditDelete)
	ON_COMMAND(ID_EDIT_DELETE, OnEditDelete)
	ON_COMMAND(ID_FILE_COMPILE, OnFileCompile)
	ON_COMMAND(ID_COMPILE_EVENCOORDINATESONLY, OnCompileEvencoordinatesonly)
	ON_UPDATE_COMMAND_UI(ID_COMPILE_EVENCOORDINATESONLY, OnUpdateCompileEvencoordinatesonly)
	ON_UPDATE_COMMAND_UI(ID_INDICATOR_FRAME, OnUpdateIndicatorFrame)
	ON_UPDATE_COMMAND_UI(ID_INDICATOR_FPS, OnUpdateIndicatorFps)
	ON_UPDATE_COMMAND_UI(ID_INDICATOR_MEDIATIME, OnUpdateIndicatorMediatime)
	ON_WM_CREATE()
	ON_WM_SIZE()
	ON_WM_HSCROLL()
	ON_COMMAND(ID_VIEW_GOTOFRAME, OnViewGotoframe)
	ON_UPDATE_COMMAND_UI(ID_VIEW_FRAMEMODE, OnUpdateViewFramemode)
	ON_UPDATE_COMMAND_UI(ID_VIEW_FIELDMODE, OnUpdateViewFieldmode)
	ON_COMMAND(ID_VIEW_FRAMEMODE, OnViewFramemode)
	ON_COMMAND(ID_VIEW_FIELDMODE, OnViewFieldmode)
	ON_UPDATE_COMMAND_UI(ID_EDIT_SELECT_ALL, OnUpdateEditSelectall)
	ON_COMMAND(ID_EDIT_SELECT_ALL, OnEditSelectall)
	ON_UPDATE_COMMAND_UI(ID_VIEW_GOTOFRAME, OnUpdateViewGotoframe)
	ON_UPDATE_COMMAND_UI(ID_EDIT_INSERTIMAGE, OnUpdateEditInsertimage)
	ON_COMMAND(ID_EDIT_INSERTIMAGE, OnEditInsertimage)
	ON_UPDATE_COMMAND_UI(ID_EDIT_REBUILDFROMADJACENTFIELDS, OnUpdateEditRebuildfromadjacentfields)
	ON_COMMAND(ID_EDIT_REBUILDFROMADJACENTFIELDS, OnEditRebuildfromadjacentfields)
	ON_COMMAND(ID_EDIT_INTERPOLATEMOTION, OnEditInterpolatemotion)
	ON_UPDATE_COMMAND_UI(ID_EDIT_INTERPOLATEMOTION, OnUpdateEditInterpolatemotion)
	ON_UPDATE_COMMAND_UI(ID_OVERLAYMODE_CHROMA, OnUpdateOverlaymodeChroma)
	ON_UPDATE_COMMAND_UI(ID_OVERLAYMODE_LUMA, OnUpdateOverlaymodeLuma)
	ON_UPDATE_COMMAND_UI(ID_OVERLAYMODE_OPAQUE, OnUpdateOverlaymodeOpaque)
	ON_COMMAND(ID_OVERLAYMODE_CHROMA, OnOverlaymodeChroma)
	ON_COMMAND(ID_OVERLAYMODE_LUMA, OnOverlaymodeLuma)
	ON_COMMAND(ID_OVERLAYMODE_OPAQUE, OnOverlaymodeOpaque)
	ON_WM_SETCURSOR()
	ON_WM_SETTINGCHANGE()
	ON_LBN_SELCHANGE(ID_FRAME_HISTORY, OnHistorySelChange)
	ON_UPDATE_COMMAND_UI(ID_Z_SENDTOBACK, OnUpdateZSendtoback)
	ON_UPDATE_COMMAND_UI(ID_Z_SENDFURTHER, OnUpdateZSendfurther)
	ON_UPDATE_COMMAND_UI(ID_Z_BRINGNEARER, OnUpdateZBringnearer)
	ON_UPDATE_COMMAND_UI(ID_Z_BRINGTOFRONT, OnUpdateZBringtofront)
	ON_COMMAND(ID_Z_SENDTOBACK, OnZSendtoback)
	ON_COMMAND(ID_Z_SENDFURTHER, OnZSendfurther)
	ON_COMMAND(ID_Z_BRINGNEARER, OnZBringnearer)
	ON_COMMAND(ID_Z_BRINGTOFRONT, OnZBringtofront)
	ON_UPDATE_COMMAND_UI(ID_EDIT_MOVEDOWN, OnUpdateEditMovedown)
	ON_UPDATE_COMMAND_UI(ID_EDIT_MOVELEFT, OnUpdateEditMoveleft)
	ON_UPDATE_COMMAND_UI(ID_EDIT_MOVERIGHT, OnUpdateEditMoveright)
	ON_UPDATE_COMMAND_UI(ID_EDIT_MOVEUP, OnUpdateEditMoveup)
	ON_UPDATE_COMMAND_UI(ID_EDIT_SIZELEFT, OnUpdateEditSizeleft)
	ON_UPDATE_COMMAND_UI(ID_EDIT_SIZERIGHT, OnUpdateEditSizeright)
	ON_UPDATE_COMMAND_UI(ID_EDIT_SIZEUP, OnUpdateEditSizeup)
	ON_UPDATE_COMMAND_UI(ID_EDIT_SIZEDOWN, OnUpdateEditSizedown)
	ON_COMMAND(ID_EDIT_MOVEDOWN, OnEditMovedown)
	ON_COMMAND(ID_EDIT_MOVELEFT, OnEditMoveleft)
	ON_COMMAND(ID_EDIT_MOVERIGHT, OnEditMoveright)
	ON_COMMAND(ID_EDIT_MOVEUP, OnEditMoveup)
	ON_COMMAND(ID_EDIT_SIZEDOWN, OnEditSizedown)
	ON_COMMAND(ID_EDIT_SIZELEFT, OnEditSizeleft)
	ON_COMMAND(ID_EDIT_SIZERIGHT, OnEditSizeright)
	ON_COMMAND(ID_EDIT_SIZEUP, OnEditSizeup)
	ON_UPDATE_COMMAND_UI(ID_EDIT_SELECTNEXT, OnUpdateEditSelectnext)
	ON_COMMAND(ID_EDIT_SELECTNEXT, OnEditSelectnext)
	ON_UPDATE_COMMAND_UI(ID_EDIT_SELECTPREVIOUS, OnUpdateEditSelectprevious)
	ON_COMMAND(ID_EDIT_SELECTPREVIOUS, OnEditSelectprevious)
	ON_COMMAND(ID_VIEW_NEXTMODIFIED, OnViewNextmodified)
	ON_COMMAND(ID_VIEW_PREVIOUSMODIFIED, OnViewPreviousmodified)
	ON_UPDATE_COMMAND_UI(ID_EDIT_SNAPSHOTFRAME, OnUpdateEditSnapshotframe)
	ON_COMMAND(ID_EDIT_SNAPSHOTFRAME, OnEditSnapshotframe)
	ON_COMMAND(ID_VIEW_INVERT, OnViewInvert)
	ON_UPDATE_COMMAND_UI(ID_VIEW_INVERT, OnUpdateViewInvert)
	ON_UPDATE_COMMAND_UI(ID_VIEW_DEINTERLACE, OnUpdateViewDeinterlace)
	ON_COMMAND(ID_VIEW_DEINTERLACE, OnViewDeinterlace)
	ON_UPDATE_COMMAND_UI(ID_VIEW_ACCURATERENDERING, OnUpdateViewAccuraterendering)
	ON_COMMAND(ID_VIEW_ACCURATERENDERING, OnViewAccuraterendering)
END_MESSAGE_MAP()

//:View maintenance

CVclipView::CVclipView()
: m_aviFile(),
	m_drawDib(),
	m_pos(0), m_order(0), m_viewMode(&frameMode), m_overlayMode(the<Overlays>()["Blend"]),
	m_dragState(new NullDrag), m_manipulator(new SelectionManipulator(m_viewMode, m_selection, CRect())),
	m_selection(0, 0, 0, 0),
	m_clipboard(),
	m_evenCoords(true), m_accurateRendering(false),
	m_invert(&identity), m_deinterlace(&identity)
{
}

CVclipView::~CVclipView()
{
}

BOOL CVclipView::PreCreateWindow(CREATESTRUCT& cs)
{
	BOOL result = CView::PreCreateWindow(cs);
	cs.dwExStyle |= WS_EX_COMPOSITED;
	return result;
}

int CVclipView::OnCreate(LPCREATESTRUCT lpCreateStruct)
{
	if (CView::OnCreate(lpCreateStruct) == -1)
		return -1;

	m_frameTrackBar.Create(WS_CHILD | WS_VISIBLE | TBS_AUTOTICKS | TBS_BOTTOM | TBS_HORZ,
		CRect(0, lpCreateStruct->cy - 20, lpCreateStruct->cx, lpCreateStruct->cy),
		this, ID_FRAME_TRACKBAR);
	m_frameTrackBar.SetRange(0, 0);

	m_historyList.Create(WS_CHILD | WS_VISIBLE | LBS_NOTIFY,
		CRect(0, lpCreateStruct->cy - 200, lpCreateStruct->cx / 2, lpCreateStruct->cy - 20),
		this, ID_FRAME_HISTORY);
	m_dependencyList.Create(WS_CHILD | WS_VISIBLE | LBS_NOTIFY,
		CRect(lpCreateStruct->cx / 2, lpCreateStruct->cy - 200, lpCreateStruct->cx / 2, lpCreateStruct->cy - 20),
		this, ID_FRAME_DEPENDENCIES);
	UpdateHistoryFont();
	
	return 0;
}

void CVclipView::OnSize(UINT nType, int cx, int cy)
{
	CView::OnSize(nType, cx, cy);

	m_frameTrackBar.MoveWindow(0, cy - 20, cx, 20);
	m_historyList.MoveWindow(0, m_aviFile.frameSize().Height(), 
		cx / 2, cy - 20 - m_aviFile.frameSize().Height());
	m_dependencyList.MoveWindow(cx / 2, m_aviFile.frameSize().Height(), 
		(cx + 1) / 2, cy - 20 - m_aviFile.frameSize().Height());
}

BOOL CVclipView::OnSetCursor(CWnd* pWnd, UINT nHitTest, UINT message)
{
	CPoint viewPoint;
	GetCursorPos(&viewPoint);
	ScreenToClient(&viewPoint);

	if (!m_viewMode->viewFromFrame(m_aviFile.frameSize()).PtInRect(viewPoint))
		return CView::OnSetCursor(pWnd, nHitTest, message);

	CPoint framePoint = m_viewMode->frameFromView(viewPoint);

	if (m_manipulator->setCursor(framePoint)) return TRUE;

	if (GetAsyncKeyState(VK_CONTROL))
	{
		SetCursor(LoadCursor(0, IDC_CROSS));
		return TRUE;
	}

	PFrameOperation op;
	GetDocument()->operations(m_pos).for_each(
		HitTest(m_viewMode, m_pos, m_order, framePoint, op) );
	if (op)
	{
		SetCursor(LoadCursor(0, IDC_ARROW));
		return TRUE;
	}
	else
	{
		SetCursor(LoadCursor(0, IDC_CROSS));
		return TRUE;
	}
}

//#View maintenance

//:Diagnostics

#ifdef _DEBUG
void CVclipView::AssertValid() const
{
	CView::AssertValid();
}

void CVclipView::Dump(CDumpContext& dc) const
{
	CView::Dump(dc);
}

CVclipDoc* CVclipView::GetDocument() const // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CVclipDoc)));
	return (CVclipDoc*)m_pDocument;
}
#endif //_DEBUG

//#Diagnostics

//:Video file

void CVclipView::OpenVideo(LPCTSTR lpszFilename)
{
	m_aviFile.open(lpszFilename);

	m_frameTrackBar.SetRange(0, m_aviFile.size());
	m_frameTrackBar.SetPageSize(m_aviFile.size() / 100);

	CRect rect;
	GetClientRect(&rect);
	OnSize(SIZE_RESTORED, rect.right, rect.bottom);

	SetFrame(0);
}

void CVclipView::CloseVideo()
{
	m_aviFile.close();

	m_frameTrackBar.SetRange(0, 0);
	m_frameTrackBar.SetPageSize(0);

	CRect rect;
	GetClientRect(&rect);
	OnSize(SIZE_RESTORED, rect.right, rect.bottom);

	SetFrame(0);
}

void CVclipView::OnViewOpenvideo()
{
	CString context = "opening video";
	try
	{
		CFileDialog openDialog(TRUE, ".avi", 0,
			OFN_ENABLESIZING | OFN_FILEMUSTEXIST | OFN_HIDEREADONLY | OFN_PATHMUSTEXIST,
			"Video files (*.avi, *.avs)|*.avi;*.avs|All files (*.*)|*.*||",
			this->GetParent());
		if (IDOK != openDialog.DoModal()) return;

		context.Format("getting full path name of %s", openDialog.GetFileName());
		TCHAR buf[MAX_PATH];
		LPTSTR filePart;
		ASSERT_NE(GetFullPathName(openDialog.GetFileName(), MAX_PATH, buf, &filePart), FALSE, "GetFullPathName");

		context.Format("opening video file %s", buf);
		GetDocument()->setSource(buf);
	}
	catch (const std::exception& e)
	{
		CString error;
		error.Format("An error occured while %s.\n%s", context, e.what());
		AfxMessageBox(error, MB_OK | MB_ICONERROR);
	}
	catch (...)
	{
		CString error;
		error.Format("An unknown error occured while %s.", context);
		AfxMessageBox(error, MB_OK | MB_ICONERROR);
	}
}

//#Video file

//:Frame seeking

void CVclipView::SetFrame(LONG frame)
{
	if (!GetDocument()->needFrame(m_pos))
	{
		m_aviFile.erase(m_pos);
	}
	
	DragCancel();
	UnselectOperation();
	
	m_pos = frame;
	m_frameTrackBar.SetPos(m_pos);
	UpdateHistory();

	Invalidate();
}

void CVclipView::OnUpdateViewNextframe(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_viewMode->hasNext(m_pos, m_order, m_aviFile.size()));
}
void CVclipView::OnViewNextframe()
{
	SetFrame(m_viewMode->next(m_pos, m_order));
}

void CVclipView::OnUpdateViewPreviousframe(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_viewMode->hasPrevious(m_pos, m_order));
}
void CVclipView::OnViewPreviousframe()
{
	SetFrame(m_viewMode->previous(m_pos, m_order));
}

void CVclipView::OnViewPreviousmodified()
{
	CVclipDoc::const_iterator previous = yk::max_element_if(GetDocument()->begin(), GetDocument()->end(),
		// op -> op.targetFrame() < m_pos
		sgi::compose1(std::bind2nd(std::less<LONG>(), m_pos), TargetFrame()),
		// a, b -> a.targetFrame() < b.targetFrame()
		yk::compose_f_gx_gy(std::less<LONG>(), TargetFrame()) );
	if (previous == GetDocument()->end()) return;
	SetFrame((*previous)->targetFrame());
}

void CVclipView::OnViewNextmodified()
{
	CVclipDoc::const_iterator next = yk::min_element_if(GetDocument()->begin(), GetDocument()->end(), 
		// op -> m_pos < op.targetFrame()
		sgi::compose1(std::bind1st(std::less<LONG>(), m_pos), TargetFrame()),
		// a, b -> a.targetFrame() < b.targetFrame()
		yk::compose_f_gx_gy(std::less<LONG>(), TargetFrame()) );
	if (next == GetDocument()->end()) return;
	SetFrame((*next)->targetFrame());
}

void CVclipView::OnUpdateViewGotoframe(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_aviFile.size());
}
void CVclipView::OnViewGotoframe()
{
	CGotoFrame gotoDialog;
	gotoDialog.SetFrame(m_pos);
	gotoDialog.SetFps(m_aviFile.millifps());
	gotoDialog.SetFrames(m_aviFile.size());
	if (IDOK == gotoDialog.DoModal())
	{
		SetFrame(gotoDialog.GetFrame());
	}
}

void CVclipView::OnHScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar)
// RATIONALE: (not using nPos)
// Trackbar sends a WM_SCROLL message.
// nPos is contained in HIWORD(wParam), which already limits positions to 16 bits.
// MFC, additionally, reinterprets it as signed short, which yields negative values
// if nPos > 32767.
{
	if (reinterpret_cast<CSliderCtrl*>(pScrollBar) == &m_frameTrackBar)
	{
		switch (nSBCode)
		{
		case TB_LINEUP:   case TB_PAGEUP:   case TB_TOP:
		case TB_LINEDOWN: case TB_PAGEDOWN: case TB_BOTTOM:
		case TB_THUMBPOSITION:
		case TB_THUMBTRACK:
		case TB_ENDTRACK:
			SetFrame(m_frameTrackBar.GetPos());
			return;
		default:
			return;
		}
	}
	else
	{
		CView::OnHScroll(nSBCode, nPos, pScrollBar);
	}
}

//#Frame seeking

//:History box

void CVclipView::UpdateHistoryFont()
{
	LOGFONT lf;
	ASSERT_NE(SystemParametersInfo(SPI_GETICONTITLELOGFONT, sizeof(lf), &lf, FALSE), 
		FALSE, "SystemParametersInfo");
	m_font.DeleteObject();
	ASSERT_NE(m_font.CreateFontIndirect(&lf),
		FALSE, "CreateFontIndirect");
	m_historyList.SetFont(&m_font);
	m_dependencyList.SetFont(&m_font);
}

void CVclipView::OnSettingChange(UINT uFlags, LPCTSTR lpszSection)
{
	CView::OnSettingChange(uFlags, lpszSection);
	if (uFlags == 0 || uFlags == SPI_SETICONTITLELOGFONT || uFlags == SPI_GETICONTITLELOGFONT)
	{
		UpdateHistoryFont();
	}
}

class ListBoxBackInserter : public std::iterator<std::output_iterator_tag, PFrameOperation, void>
{
private:
	CListBox* m_pBox;
public:
	explicit ListBoxBackInserter(CListBox& box)
		: m_pBox(&box)
	{}

	const ListBoxBackInserter& operator*() const
	{
		return *this;
	}

	void operator=(PFrameOperation op) const
	{
		std::ostringstream oss;
		oss << op;
		m_pBox->SetItemDataPtr(
			m_pBox->AddString(oss.str().c_str()), 
			op.get());
	}

	ListBoxBackInserter& operator++()
	{
		return *this;
	}

	const ListBoxBackInserter& operator++(int)
	{
		return *this;
	}
};

void CVclipView::UpdateHistory()
{
	m_historyList.ResetContent();
	CVclipDoc::range operations = GetDocument()->operations(m_pos);

	std::remove_copy_if(operations.first, operations.second,
		ListBoxBackInserter(m_historyList),
		std::not1(
			sgi::compose1(std::bind2nd(std::equal_to<LONG>(), m_pos), TargetFrame())
		)
	);
}

void CVclipView::OnHistorySelChange()
{
	int sel = m_historyList.GetCurSel();
	if (sel < 0) return;
	SelectOperation(PFrameOperation(
		static_cast<FrameOperation*>(m_historyList.GetItemDataPtr(sel))));
}

void CVclipView::UpdateDependencies()
{
	m_dependencyList.ResetContent();
	CVclipDoc::range operations = GetDocument()->operations(m_pos);

	std::remove_copy_if(operations.first, operations.second,
		ListBoxBackInserter(m_dependencyList),
		std::not1(
			std::bind1st(DependencyExists(), m_manipulator->operation())
		)
	);

	m_dependencyList.SetCurSel(IndexOfDataPtr(m_dependencyList, m_manipulator->operation().get()));
}

//#History box

//:Overlay mode

class UpdateOverlayMode : public IOperationVisitor
{
private:
	HOverlay* m_pMode;
public:
	explicit UpdateOverlayMode(HOverlay* pMode) : m_pMode(pMode) {}

	void operator()(FramePatch& patch) { *m_pMode = patch.mode(); }
	void operator()(FieldPatch& patch) { *m_pMode = patch.mode(); }
	void operator()(FrameImage& /*image*/) {}
	void operator()(FieldRebuild& /*rebuild*/) {}
	void operator()(MotionInterpolate& /*inter*/) {}
};

class SetOverlayMode : public IOperationVisitor
{
private:
	HOverlay m_mode;
public:
	explicit SetOverlayMode(HOverlay mode) : m_mode(mode) {}

	void operator()(FramePatch& patch) { patch.setMode(m_mode); }
	void operator()(FieldPatch& patch) { patch.setMode(m_mode); }
	void operator()(FrameImage& /*image*/) {}
	void operator()(FieldRebuild& /*rebuild*/) {}
	void operator()(MotionInterpolate& /*inter*/) {}
};

void CVclipView::UpdateOverlaymode(CCmdUI* pCmdUI, HOverlay mode) const
{
	pCmdUI->Enable(m_aviFile.size());
	pCmdUI->SetRadio(m_overlayMode == mode);
}
void CVclipView::SetOverlaymode(HOverlay mode)
{
	m_manipulator->accept(SetOverlayMode(mode));
	m_overlayMode = mode;
}

void CVclipView::OnUpdateOverlaymodeOpaque(CCmdUI *pCmdUI)
{
	UpdateOverlaymode(pCmdUI, the<Overlays>()["Blend"]);
}
void CVclipView::OnOverlaymodeOpaque()
{
	SetOverlaymode(the<Overlays>()["Blend"]);
}

void CVclipView::OnUpdateOverlaymodeChroma(CCmdUI *pCmdUI)
{
	UpdateOverlaymode(pCmdUI, the<Overlays>()["Chroma"]);
}
void CVclipView::OnOverlaymodeChroma()
{
	SetOverlaymode(the<Overlays>()["Chroma"]);
}

void CVclipView::OnUpdateOverlaymodeLuma(CCmdUI *pCmdUI)
{
	UpdateOverlaymode(pCmdUI, the<Overlays>()["Luma"]);
}
void CVclipView::OnOverlaymodeLuma()
{
	SetOverlaymode(the<Overlays>()["Luma"]);
}

//#Overlay mode

//:Z-order

void CVclipView::OnUpdateZSendtoback(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_manipulator->isOperation());
}
void CVclipView::OnZSendtoback()
{
	GetDocument()->sendToBack(m_manipulator->operation());
}

void CVclipView::OnUpdateZSendfurther(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_manipulator->isOperation() && 
		GetDocument()->canSendFurther(m_manipulator->operation()));
}
void CVclipView::OnZSendfurther()
{
	GetDocument()->sendFurther(m_manipulator->operation());
}

void CVclipView::OnUpdateZBringnearer(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_manipulator->isOperation() &&
		GetDocument()->canBringNearer(m_manipulator->operation()));
}
void CVclipView::OnZBringnearer()
{
	GetDocument()->bringNearer(m_manipulator->operation());
}

void CVclipView::OnUpdateZBringtofront(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_manipulator->isOperation());
}
void CVclipView::OnZBringtofront()
{
	GetDocument()->bringToFront(m_manipulator->operation());
}

//#Z-order

//:Compilation

void CVclipView::OnUpdateCompileEvencoordinatesonly(CCmdUI *pCmdUI)
{
	pCmdUI->SetCheck(m_evenCoords);
}
void CVclipView::OnCompileEvencoordinatesonly()
{
	m_evenCoords = !m_evenCoords;
}

void CVclipView::OnFileCompile()
{
	YK_TRACED("OnFileCompile");

	CFileDialog saveDialog(FALSE, ".avs", 0, OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,
		"AviSynth Files (*.avs)|*.avs|All Files (*.*)|*.*||", GetParent());
	if (IDOK == saveDialog.DoModal())
	{
		GetDocument()->compile(saveDialog.GetFileName(), m_evenCoords, 
			m_aviFile.size(), m_aviFile.frameSize());
	}
}

//#Compilation

//:Document

void CVclipView::OnUpdate(CView* /*pSender*/, LPARAM lHint, CObject* /*pHint*/)
{
	CString context = "updating";
	try
	{
		if (lHint == CVclipDoc::updateSource || lHint == CVclipDoc::updateWhole)
		{
			// source changed
			CloseVideo();

			if (!GetDocument()->source().empty())
			{
				context.Format("opening video file %s", GetDocument()->source().c_str());
				OpenVideo(GetDocument()->source().c_str());
			}
		}
		if (lHint == m_pos || lHint == CVclipDoc::updateWhole)
		{
			UpdateHistory();
			m_historyList.SetCurSel(IndexOfDataPtr(m_historyList, m_manipulator->operation().get()));
			Invalidate();
		}
	}
	catch (const std::exception& e)
	{
		CString error;
		error.Format("An error occured while %s.\n%s", context, e.what());
		AfxMessageBox(error, MB_OK | MB_ICONERROR);
	}
	catch (...)
	{
		CString error;
		error.Format("An unknown error occured while %s.", context);
		AfxMessageBox(error, MB_OK | MB_ICONERROR);
	}
}

//:Status bar

void CVclipView::OnUpdateIndicatorFrame(CCmdUI *pCmdUI)
{
	pCmdUI->Enable();
	pCmdUI->SetText(m_viewMode->formatPosition(m_pos, m_order, m_aviFile.size()));
}

void CVclipView::OnUpdateIndicatorMediatime(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_aviFile.size() != 0);
	if (m_aviFile.size())
		pCmdUI->SetText(formatMediaTime(mediaTimeFromFrame(m_pos, m_aviFile.millifps())));
}

void CVclipView::OnUpdateIndicatorFps(CCmdUI *pCmdUI)
{
	pCmdUI->Enable();
	CString text;
	text.Format("%d.%03d fps", m_aviFile.millifps() / 1000, m_aviFile.millifps() % 1000);
	pCmdUI->SetText(text);
}

//#Status bar

/////
