// vclipView.h : interface of the CVclipView class
//
#pragma once

#include <vfw.h>
#include <map>
#include "FrameCache.h"
#include "DrawDib.h"
#include "FrameOperation.h"
#include "vclipDoc.h"
#include "OverlayMode.h"
#include "ViewMode.h"

class IDragState
{
public:
	virtual ~IDragState() {}
	virtual void clipCursor(/*by value*/ CRect rect) const = 0;
	virtual bool mouseMove(const CPoint& point) const = 0;
	virtual void mouseUp(const CPoint& point) const = 0;
	virtual bool canCancel() const = 0;
	virtual void dragCancel() const = 0;
};

class IManipulator
{
public:
	virtual ~IManipulator() {}
	virtual void draw(CDC* pDC, const ViewMode* mode) const = 0;
	virtual bool wantClick(const CPoint& point) const = 0;
	virtual std::auto_ptr<IDragState> mouseDown(const CPoint& point) const = 0;
	virtual bool canErase() const = 0;
	virtual void erase(CVclipDoc& document) const = 0;
	virtual void accept(IOperationVisitor& visitor) const = 0;
	virtual bool setCursor(const CPoint& point) const = 0;
	virtual bool isOperation() const = 0;
	virtual PFrameOperation operation() const = 0;
	virtual CRect getRect() const = 0;
	virtual void setRect(const CRect& rect) const = 0;
	virtual void deflateRect(CRect& rect, const CRect& offset) const = 0;
	virtual bool canMove(int dx, int dy) const = 0;
	virtual void moveBy(int dx, int dy) const = 0;
	virtual bool canSize(int dx, int dy) const = 0;
	virtual void sizeBy(int dx, int dy) const = 0;
};

class IPostProcessor
{
public:
	virtual void operator()(PFrame frame) const = 0;
	virtual const IPostProcessor* select(const IPostProcessor* identity, const IPostProcessor* worker) const = 0;
};

class CVclipView : public CView
{
protected: // create from serialization only
	CVclipView();
	DECLARE_DYNCREATE(CVclipView)

private: // Video file
	AviFile m_aviFile;
	void OpenVideo(LPCTSTR lpszFilename);
	void CloseVideo();

private: // Image cache
	ImageCache m_imageCache;

private: // DIB drawing
	DrawDib m_drawDib;

private: // frame seeking
	LONG m_pos;
	LONG m_order;
	CSliderCtrl m_frameTrackBar;
	void SetFrame(LONG frame);

private: // history
	CFont m_font;
	CListBox m_historyList;
	void UpdateHistoryFont();
	void UpdateHistory();

private: // dependencies
	CListBox m_dependencyList;
	void UpdateDependencies();

private: // selection manipulation
	CRect m_selection;
	std::auto_ptr<IDragState> m_dragState;
	void DragCancel();
	void EndDrag(CPoint point);
	std::auto_ptr<IManipulator> m_manipulator;
	void SelectOperation(PFrameOperation op);
	void UnselectOperation();

private: // clipboard
	Clipboard m_clipboard;

private: // view mode
	ViewMode* m_viewMode;

private: // overlay mode
	HOverlay m_overlayMode;

private: // postprocessors
	const IPostProcessor* m_invert;
	const IPostProcessor* m_deinterlace;
	bool m_accurateRendering;

// Attributes
private:

	bool m_evenCoords;
public:
	CVclipDoc* GetDocument() const;

// Operations
public:

// Overrides
private:
	bool NeedExternalRender(LONG frame) const;
	PFrame PrepareFrame(CDC* pDC) const;
public:
	virtual void OnDraw(CDC* pDC);  // overridden to draw this view
	virtual BOOL PreCreateWindow(CREATESTRUCT& cs);
protected:

// Implementation
public:
	virtual ~CVclipView();
#ifdef _DEBUG
	virtual void AssertValid() const;
	virtual void Dump(CDumpContext& dc) const;
#endif

protected:

// Generated message map functions
private:
	void UpdateOverlaymode(CCmdUI* pCmdUI, HOverlay mode) const;
	void SetOverlaymode(HOverlay mode);
protected:
	DECLARE_MESSAGE_MAP()
public:
	afx_msg void OnViewOpenvideo();
	afx_msg void OnViewNextframe();
	afx_msg void OnUpdateViewNextframe(CCmdUI *pCmdUI);
	afx_msg void OnUpdateViewPreviousframe(CCmdUI *pCmdUI);
	afx_msg void OnViewPreviousframe();
	afx_msg void OnLButtonDown(UINT nFlags, CPoint point);
	afx_msg void OnLButtonUp(UINT nFlags, CPoint point);
	afx_msg void OnMouseMove(UINT nFlags, CPoint point);
	afx_msg void OnUpdateEditCopy(CCmdUI *pCmdUI);
	afx_msg void OnEditCopy();
	afx_msg void OnUpdateEditPaste(CCmdUI *pCmdUI);
	afx_msg void OnEditPaste();
	afx_msg void OnDragcancel();
	afx_msg void OnUpdateEditDelete(CCmdUI *pCmdUI);
	afx_msg void OnEditDelete();
	afx_msg void OnFileCompile();
protected:
	virtual void OnUpdate(CView* /*pSender*/, LPARAM /*lHint*/, CObject* /*pHint*/);
public:
	afx_msg void OnCompileEvencoordinatesonly();
	afx_msg void OnUpdateCompileEvencoordinatesonly(CCmdUI *pCmdUI);
	afx_msg void OnUpdateIndicatorFrame(CCmdUI *pCmdUI);
	afx_msg void OnUpdateIndicatorFps(CCmdUI *pCmdUI);
	afx_msg void OnUpdateIndicatorMediatime(CCmdUI *pCmdUI);
	afx_msg int OnCreate(LPCREATESTRUCT lpCreateStruct);
	afx_msg void OnSize(UINT nType, int cx, int cy);
	afx_msg void OnHScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar);
	afx_msg void OnViewGotoframe();
	afx_msg void OnUpdateViewFramemode(CCmdUI *pCmdUI);
	afx_msg void OnUpdateViewFieldmode(CCmdUI *pCmdUI);
	afx_msg void OnViewFramemode();
	afx_msg void OnViewFieldmode();
	afx_msg void OnUpdateEditSelectall(CCmdUI *pCmdUI);
	afx_msg void OnEditSelectall();
	afx_msg void OnUpdateViewGotoframe(CCmdUI *pCmdUI);
	afx_msg void OnUpdateEditInsertimage(CCmdUI *pCmdUI);
	afx_msg void OnEditInsertimage();
	afx_msg void OnUpdateEditRebuildfromadjacentfields(CCmdUI *pCmdUI);
	afx_msg void OnEditRebuildfromadjacentfields();
	afx_msg void OnEditInterpolatemotion();
	afx_msg void OnUpdateEditInterpolatemotion(CCmdUI *pCmdUI);
	afx_msg void OnUpdateOverlaymodeChroma(CCmdUI *pCmdUI);
	afx_msg void OnUpdateOverlaymodeLuma(CCmdUI *pCmdUI);
	afx_msg void OnUpdateOverlaymodeOpaque(CCmdUI *pCmdUI);
	afx_msg void OnOverlaymodeChroma();
	afx_msg void OnOverlaymodeLuma();
	afx_msg void OnOverlaymodeOpaque();
	afx_msg BOOL OnSetCursor(CWnd* pWnd, UINT nHitTest, UINT message);
	afx_msg void OnSettingChange(UINT uFlags, LPCTSTR lpszSection);
	afx_msg void OnHistorySelChange();
	afx_msg void OnUpdateZSendtoback(CCmdUI *pCmdUI);
	afx_msg void OnUpdateZSendfurther(CCmdUI *pCmdUI);
	afx_msg void OnUpdateZBringnearer(CCmdUI *pCmdUI);
	afx_msg void OnUpdateZBringtofront(CCmdUI *pCmdUI);
	afx_msg void OnZSendtoback();
	afx_msg void OnZSendfurther();
	afx_msg void OnZBringnearer();
	afx_msg void OnZBringtofront();
	afx_msg void OnUpdateEditMovedown(CCmdUI *pCmdUI);
	afx_msg void OnUpdateEditMoveleft(CCmdUI *pCmdUI);
	afx_msg void OnUpdateEditMoveright(CCmdUI *pCmdUI);
	afx_msg void OnUpdateEditMoveup(CCmdUI *pCmdUI);
	afx_msg void OnUpdateEditSizeleft(CCmdUI *pCmdUI);
	afx_msg void OnUpdateEditSizeright(CCmdUI *pCmdUI);
	afx_msg void OnUpdateEditSizeup(CCmdUI *pCmdUI);
	afx_msg void OnUpdateEditSizedown(CCmdUI *pCmdUI);
	afx_msg void OnEditMovedown();
	afx_msg void OnEditMoveleft();
	afx_msg void OnEditMoveright();
	afx_msg void OnEditMoveup();
	afx_msg void OnEditSizedown();
	afx_msg void OnEditSizeleft();
	afx_msg void OnEditSizeright();
	afx_msg void OnEditSizeup();
	afx_msg void OnUpdateEditSelectnext(CCmdUI *pCmdUI);
	afx_msg void OnEditSelectnext();
	afx_msg void OnUpdateEditSelectprevious(CCmdUI *pCmdUI);
	afx_msg void OnEditSelectprevious();
	afx_msg void OnViewNextmodified();
	afx_msg void OnViewPreviousmodified();
	afx_msg void OnUpdateEditSnapshotframe(CCmdUI *pCmdUI);
	afx_msg void OnEditSnapshotframe();
	afx_msg void OnViewInvert();
	afx_msg void OnUpdateViewInvert(CCmdUI *pCmdUI);
	afx_msg void OnUpdateViewDeinterlace(CCmdUI *pCmdUI);
	afx_msg void OnViewDeinterlace();
	afx_msg void OnUpdateViewAccuraterendering(CCmdUI *pCmdUI);
	afx_msg void OnViewAccuraterendering();
};

#ifndef _DEBUG  // debug version in vclipView.cpp
inline CVclipDoc* CVclipView::GetDocument() const
   { return reinterpret_cast<CVclipDoc*>(m_pDocument); }
#endif

